#!/usr/bin/env python
# -*- coding: utf-8 -*-
from dataclasses import dataclass
from functools import partial
from typing import Callable

import pytest

from registry import RegisteredFunction, Registry


@dataclass
class CallableClass:
    foo: int = 1

    def __call__(self, *args, **kwargs):
        ...


@dataclass
class CallableClassWithArg:
    foo: int = 1

    def __call__(self, x: int):
        return x


@dataclass(kw_only=True)
class KeywordOnlyClass:
    foo: int = 1

    def __call__(self, x: int):
        return x


def dummy_func(*args, **kwargs):
    ...


class TestRegisteredFunction:
    def test_call(self):
        fn = CallableClassWithArg()
        func = RegisteredFunction(fn, "foo")
        assert func(x=1) == 1

    @pytest.mark.parametrize(
        "fn,exp",
        [
            pytest.param(dummy_func, False),
            pytest.param(str, True),
            pytest.param(CallableClass, True),
            pytest.param(CallableClass(), False),
            pytest.param(KeywordOnlyClass, True),
        ],
    )
    def test_is_type(self, fn, exp):
        func = RegisteredFunction(fn, "foo")
        assert func.is_type == exp

    @pytest.mark.parametrize(
        "fn,exp",
        [
            pytest.param(dummy_func, dummy_func),
            pytest.param(CallableClass, CallableClass()),
            pytest.param(CallableClass(), CallableClass()),
            pytest.param(KeywordOnlyClass, KeywordOnlyClass()),
            pytest.param(KeywordOnlyClass(), KeywordOnlyClass()),
        ],
    )
    def test_instantiate(self, fn, exp):
        func = RegisteredFunction(fn, "foo")
        inst = func.instantiate(bar="bar")
        assert inst.fn == exp

    @pytest.mark.parametrize(
        "fn,exp",
        [
            pytest.param(dummy_func, dummy_func),
            pytest.param(CallableClass, CallableClass(foo=2)),
            pytest.param(CallableClass(), CallableClass()),
            pytest.param(partial(CallableClass, foo=2), CallableClass(foo=2)),
            pytest.param(KeywordOnlyClass, KeywordOnlyClass(foo=2)),
            pytest.param(partial(KeywordOnlyClass, foo=2), KeywordOnlyClass(foo=2)),
        ],
    )
    def test_instantiate_with_metadata(self, fn, exp):
        func = RegisteredFunction(fn, "foo", {"foo": 2})
        inst = func.instantiate_with_metadata(bar="bar")
        assert inst.fn == exp

    def test_cast(self):
        fn = CallableClass()
        func = RegisteredFunction(fn, "foo")
        casted_func = func.cast(Callable[[int], int])
        assert casted_func == func

    def test_bind(self):
        fn = CallableClassWithArg()
        func = RegisteredFunction(fn, "foo")
        bound = func.bind(x=1, y=2)
        assert bound() == 1

    def test_bind_with_metadata(self):
        fn = CallableClassWithArg()
        func = RegisteredFunction(fn, "foo", {"x": 1})
        bound = func.bind_metadata(y=2)
        assert bound() == 1


class TestRegistry:
    @pytest.mark.parametrize("name", ["name1", "name2"])
    def test_construct(self, name):
        reg = Registry(name)
        assert reg.name == name
        assert len(reg) == 0

    def test_repr(self):
        reg = Registry("name")
        s = str(reg)
        assert isinstance(s, str)

    @pytest.mark.parametrize("func_name", ["func1", "func2"])
    def test_register_class(self, func_name):
        reg = Registry("name")
        assert reg.name == "name"

        @reg(name=func_name)
        class DummyClass:
            pass

        assert func_name in reg
        assert reg.get(func_name).fn is DummyClass

    @pytest.mark.parametrize("func_name", ["func1", "func2"])
    def test_register_function(self, func_name):
        reg = Registry("name")

        def func(x: int) -> None:
            pass

        reg(func, name=func_name)
        assert func_name in reg
        assert reg.get(func_name).fn is func

    @pytest.mark.parametrize("func_name", ["func1", "func2"])
    def test_contains(self, func_name):
        reg = Registry("name")

        @reg(name=func_name)
        class DummyClass:
            pass

        assert func_name in reg

    @pytest.mark.parametrize("length", [0, 1, 2])
    def test_length(self, length):
        reg = Registry("name")

        def func():
            pass

        for i in range(length):
            reg(func, name=f"func-{i}")
        assert len(reg) == length

    @pytest.mark.parametrize(
        "names",
        [
            ("f1",),
            ("f1", "f2"),
        ],
    )
    def test_available_keys(self, names):
        reg = Registry("name")

        def func():
            pass

        for name in names:
            reg(func, name=name)
        assert reg.available_keys() == sorted(names)

    def test_type_annotation(self):
        def func(x: int, y: int) -> float:
            return float(x + y)

        reg = Registry("name", bound=Callable[[int, int], float])
        reg(func, name="func")
        x = reg.get("func")
        assert x(1, 2) == float(3)

    @pytest.mark.parametrize(
        "inp",
        [
            {"noop": lambda x: x},
            {"add1": lambda x: x + 1},
            {"noop": lambda x: x, "add1": lambda x: x + 1},
        ],
    )
    def test_register_dict(self, inp):
        reg = Registry("name")
        reg.register_dict(inp)

        for k, v in inp.items():
            assert k in reg
            assert reg.get(k).fn is v

    @pytest.mark.parametrize(
        "override",
        [
            True,
            pytest.param(False, marks=pytest.mark.xfail(raises=RuntimeError, strict=True)),
        ],
    )
    def test_override(self, override):
        reg = Registry("name")
        reg(lambda x: x, name="func")
        reg(lambda x: x + 1, name="func", override=override)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import importlib.metadata

from .registry import RegisteredFunction, Registry, bind_relevant_kwargs, has_var_kwargs, iterate_arg_names


__version__ = importlib.metadata.version("registry")

__all__ = ["Registry", "RegisteredFunction", "iterate_arg_names", "has_var_kwargs", "bind_relevant_kwargs"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from dataclasses import dataclass, field, replace
from functools import partial
from inspect import signature
from typing import Any, Callable, Dict, Generic, Iterator, List, Optional, Type, TypeVar, Union, cast


try:
    from typing import ParamSpec  # pragma: no cover
except ImportError:
    from typing_extensions import ParamSpec  # pragma: no cover

try:
    from typing import Self  # pragma: no cover
except ImportError:
    from typing_extensions import Self  # pragma: no cover


T = TypeVar("T")
U = TypeVar("U")
F = TypeVar("F", bound="RegisteredFunction")
P = ParamSpec("P")

P_cast = ParamSpec("P_cast")
T_cast = TypeVar("T_cast")

C = Union[Callable[P, T], Type[Callable[P, T]]]

_REGISTERED_FUNCTION = Dict[str, Any]


def iterate_arg_names(func: Callable) -> Iterator[str]:
    r"""Iterates over the arg names of a callable"""
    sig = signature(func)
    for name, param in sig.parameters.items():
        if param.kind in (param.POSITIONAL_OR_KEYWORD, param.KEYWORD_ONLY):
            yield name


def has_var_kwargs(func: Callable) -> bool:
    r"""Check if a callable supports **kwargs"""
    sig = signature(func)
    return any(p.kind == p.VAR_KEYWORD for p in sig.parameters.values())


def bind_relevant_kwargs(func: Callable[..., T], **kwargs) -> Callable[..., T]:
    r"""Binds keyword args to a callable"""
    if not kwargs:
        return func

    # edge case to handle signature hidden in a functools.partial
    if isinstance(func, partial):
        return bind_relevant_kwargs(func.func, **{**func.keywords, **kwargs})

    # for a function with **kwargs, we can bind everything
    if has_var_kwargs(func):
        return partial(func, **kwargs)

    # for functions without **kwargs, only bind kwargs that appear in function signature
    arg_names = set(iterate_arg_names(func))
    kwargs_to_bind = {k: v for k, v in kwargs.items() if k in arg_names}
    return partial(func, **kwargs_to_bind)


@dataclass(frozen=True)
class RegisteredFunction(Generic[P, T]):
    fn: Callable[P, T]
    name: str
    metadata: Dict[str, Any] = field(default_factory=dict)

    @property
    def is_type(self) -> bool:
        r"""Checks if ``fn`` is a type"""
        if isinstance(self.fn, partial):
            return isinstance(self.fn.func, type)
        return isinstance(self.fn, type)

    def instantiate(self, **kwargs) -> "RegisteredFunction":
        if self.is_type:
            fn = bind_relevant_kwargs(self.fn, **kwargs)()
            return replace(self, fn=fn)
        return self

    def instantiate_with_metadata(self, **kwargs) -> "RegisteredFunction":
        kwargs = {**self.metadata, **kwargs}
        return self.instantiate(**kwargs)

    def cast(self, typ: Type[Callable[P_cast, T_cast]]) -> "RegisteredFunction[P_cast, T_cast]":
        return self  # type: ignore

    def bind(self: Self, **kwargs) -> "RegisteredFunction":
        return replace(self, fn=bind_relevant_kwargs(self.fn, **kwargs))

    def bind_metadata(self: Self, **kwargs) -> "RegisteredFunction":
        kwargs = {**self.metadata, **kwargs}
        return self.bind(**kwargs)

    def __call__(self: Self, *args: P.args, **kwargs: P.kwargs) -> T:
        return self.fn(*args, **kwargs)


class Registry(Generic[P, T]):
    """A registry is used to register callables and associated metadata under a string name for easy access.

    Modeled after https://github.com/PyTorchLightning/lightning-flash/flash/core/registry.py

    Args:
        name: Name of the registry
        bind_metadata: If ``True``, treat metadata as keyword args that will be bound using :func:`functools.partial`
        bound: A callable from which to bind generic type variables. Used only for type checking.
    """

    def __init__(
        self,
        name: str,
        bound: Optional[Type[Callable[P, T]]] = None,
    ) -> None:
        self.name = name
        self.functions: Dict[str, RegisteredFunction[P, T]] = {}

    def __len__(self) -> int:
        return len(self.functions)

    def __contains__(self, key: Any) -> bool:
        return any(key == e.name for e in self.functions.values())

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name}, " f"functions={self.functions.keys()})"

    def get(self, key: Union[str, Callable[P, T]]) -> RegisteredFunction[P, T]:
        if isinstance(key, str):
            if key not in self:
                raise KeyError(f"Key: {key} is not in {type(self).__name__}. Available keys: {self.available_keys()}")
            result = self.functions[key]
        elif callable(key):
            result = RegisteredFunction(key, "", {})
        else:
            raise TypeError(f"`key` should be str or callable, found {type(key)}")  # pragma: no cover
        return cast(RegisteredFunction[P, T], result)

    def remove(self, key: str) -> None:
        self.functions.pop(key)

    def _register_function(
        self,
        fn: Callable[P, T],
        name: Optional[str] = None,
        override: bool = False,
        metadata: Optional[Dict[str, Any]] = None,
    ):
        if not callable(fn):
            raise TypeError(f"You can only register a callable, found: {fn}")  # pragma: no cover

        if name is None:
            if hasattr(fn, "func"):
                name = fn.func.__name__
            else:
                name = fn.__name__
        assert isinstance(name, str)

        item: RegisteredFunction[P, T] = RegisteredFunction(fn, name, metadata or {})  # type: ignore

        if not override and name in self:
            raise RuntimeError(  # pragma: no cover
                f"Function with name: {name} and metadata: {metadata} is already present within {self}."
                " HINT: Use `override=True`."
            )

        self.functions[name] = item

    def __call__(
        self,
        fn: Optional[Callable[P, T]] = None,
        name: Optional[str] = None,
        override: bool = False,
        **metadata,
    ) -> Callable:
        """This function is used to register new functions to the registry along their metadata.
        Functions can be filtered using metadata using the ``get`` function.
        """
        if fn is not None:
            self._register_function(fn=fn, name=name, override=override, metadata=metadata)
            return fn

        # raise the error ahead of time
        if not (name is None or isinstance(name, str)):
            raise TypeError(f"`name` must be a str, found {name}")  # pragma: no cover

        def _register(cls):
            self._register_function(fn=cls, name=name, override=override, metadata=metadata)
            return cls

        return _register

    def available_keys(self) -> List[str]:
        return sorted(self.functions.keys())

    def register_dict(self, fn_dict: Dict[str, Callable[P, T]], override: bool = False) -> None:
        r"""Register a dictionary of functions. The keys of the dictionary will be used as the name of the function.

        Args:
            fn_dict: A dictionary of functions to register.
            override: If ``True``, override existing functions.
        """
        for name, fn in fn_dict.items():
            self(fn=fn, name=name, override=override)
